import json
import os
import pathlib
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq


def local_file2dataframe(file: Union[str, pathlib.Path]):
    """Read the dataframe of tortilla file given a local path.

    Args:
        files (Union[str, pathlib.Path]): A local path pointing to the
            tortilla file.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """
    with open(file, "rb") as f:
        static_bytes = f.read(18)

        # Extract metadata
        MB, FO, FL = static_bytes[:2], static_bytes[2:10], static_bytes[10:18]
        if MB not in {b"#y", b"WX"}:
            raise ValueError(
                "Invalid file type: must be either a Tortilla 🫓 or a TACO 🌮"
            )

        footer_offset = int.from_bytes(FO, "little")
        footer_length = int.from_bytes(FL, "little")

        # Read the footer
        f.seek(footer_offset)
        dataframe = pq.read_table(pa.BufferReader(f.read(footer_length))).to_pandas()

    # Add auxiliary columns
    dataframe["internal:mode"] = "local"
    dataframe["internal:subfile"] = dataframe.apply(
        lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},{file}",
        axis=1,
    )
    return dataframe


def local_files2dataframe(files: Union[List[str], List[pathlib.Path]]) -> pd.DataFrame:
    """Read the dataframe of tortilla files given local paths.

    Args:
        files (Union[List[str], List[Path]]): A list of local
            paths pointing to the tortilla files.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """
    max_workers = len(files) if len(files) < os.cpu_count() else os.cpu_count()
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(local_file2dataframe, url) for url in files]
        results = []
        for future in as_completed(futures):
            result = future.result()
            if result is not None:
                results.append(result)
    return pd.concat(results, ignore_index=True)


def local_lazyfile2dataframe(
    offset: int, file: Union[str, pathlib.Path]
) -> pd.DataFrame:
    """Read the dataframe of tortilla file that is a subfile
    of a larger file.

    Args:
        offset (int): The offset of the subfile.
        file (Union[str, pathlib.Path]): A local path pointing to the
            main tortilla file.

    Returns:
        pd.DataFrame: The dataframe of the tortilla file.
    """

    with open(file, "rb") as f:
        # Seek to the OFFSET
        f.seek(offset)

        static_bytes = f.read(18)

        # SPLIT the static bytes
        MB: bytes = static_bytes[:2]
        FO: bytes = static_bytes[2:10]
        FL: bytes = static_bytes[10:18]
        # DP: str = static_bytes[42:50]

        if MB not in {b"#y", b"WX"}:
            raise ValueError(
                "Invalid file type: must be either a Tortilla 🫓 or a TACO 🌮"
            )

        # Read the NEXT 8 bytes of the file
        footer_offset: int = int.from_bytes(FO, "little") + offset

        # Seek to the FOOTER offset
        f.seek(footer_offset)

        # Select the FOOTER length
        # Read the FOOTER
        footer_length: int = int.from_bytes(FL, "little")
        dataframe = pq.read_table(pa.BufferReader(f.read(footer_length))).to_pandas()

        # Fix the offset
        dataframe["tortilla:offset"] = dataframe["tortilla:offset"] + offset

        # Convert dataset to DataFrame
        dataframe["internal:mode"] = "local"
        dataframe["internal:subfile"] = dataframe.apply(
            lambda row: f"/vsisubfile/{row['tortilla:offset']}_{row['tortilla:length']},{file}",
            axis=1,
        )

    return dataframe


def local_file2metadata(file: Union[str, pathlib.Path]) -> dict:
    """Read the dataframe of a taco file given a local path.

    Args:
        file (Union[str, pathlib.Path]): A local path pointing to the
            taco file.

    Returns:
        dict: The metadata of the taco file.
    """
    with open(file, "rb") as f:
        f.seek(26)

        # Read the Collection offset (CO)
        CO: int = int.from_bytes(f.read(8), "little")

        # Read the Collection length (CL)
        CL: int = int.from_bytes(f.read(8), "little")

        # Seek to the Collection offset
        f.seek(CO)

        # Read the Collection (JSON UTF-8 encoded)
        collection: dict = json.loads(f.read(CL).decode())

    return collection


def local_files2metadata(files: Union[List[str], List[pathlib.Path]]) -> dict:
    """Read the metadata of taco files given local paths.

    Args:
        files (Union[List[str], List[pathlib.Path]]): A list of local
            paths pointing to the taco files.

    Returns:
        dict: The metadata of the taco file.
    """

    return local_file2dataframe(files[0])



import pathlib
import re
import urllib
from typing import List, Tuple, Union

import geopandas as gpd
import requests
import shapely.wkt


def is_valid_url(url: Union[str, List[str]]) -> bool:
    """Check if a URL or list of URLs is valid.

    Args:
        url (Union[str, List[str]]): The URL(s) to check. It can
            be a single URL or a list of URLs.

    Returns:
        bool: True if all URLs are valid, False otherwise.
    """
    if isinstance(url, pathlib.Path):
        return False

    if isinstance(url, list):
        return all(is_valid_url(single_url) for single_url in url)

    try:
        result = urllib.parse.urlparse(url)
        return all([result.scheme in ["http", "https"], result.netloc])
    except ValueError:
        return False


def split_name_and_path(file: Union[str, pathlib.Path]) -> Tuple[str, str]:
    """
    Split a file path or URL into its name and path components.

    Args:
        file (Union[str, pathlib.Path]): The input file or URL.

    Returns:
        Tuple[str, str]: A tuple containing the name and path.
    """
    if isinstance(file, (str, pathlib.Path)):
        # Handle URLs
        parsed = urllib.parse.urlparse(file)
        if parsed.scheme in {"http", "https"}:
            # Extract the name and path from a URL
            name = pathlib.PurePosixPath(parsed.path).name
            path = f"{parsed.scheme}://{parsed.netloc}{pathlib.PurePosixPath(parsed.path).parent.as_posix()}"
        else:
            # Handle local file paths
            file_path = pathlib.Path(file)
            name = file_path.name
            path = file_path.parent.as_posix()
        return name, path
    else:
        raise ValueError("Input must be a string or pathlib.Path.")


def snippet2files(
    file: Union[str, pathlib.Path, List[str], List[pathlib.Path]]
) -> Union[List[pathlib.Path], List[str], str, pathlib.Path]:
    """Convert snippets of a multi-part file to a list of files.

    Args:
        file (Union[str, pathlib.Path, List[str], List[pathlib.Path]]): A file, a
            list of files, or a snippet of a multi-part file.

    Raises:
        FileNotFoundError: If a file is not found.
        FileNotFoundError: If a partial file is missing.

    Returns:
        List[pathlib.Path]: A list of files or a single path. The path can be a
            local path or a URL.
    """

    # Check if the file is a list
    if isinstance(file, list):
        files = file
    else:
        # Does the file finish with: *.tortilla?
        if re.match(r".*\*\.(tortilla|taco)$", str(file)):

            # Split in name and path
            name, path = split_name_and_path(file)

            # Get the filename without the snippet (i.e., *.tortilla)
            filename: str = re.sub(r"\*\.(tortilla|taco)$", "", name)

            # get suffix
            path_suffix: str = pathlib.Path(file).suffix

            # check if file is a url
            if is_valid_url(path):
                # It is expected that the file is a multi-part file in the same url path
                dumbfile: str = f"{path}/{filename}.0000.part{path_suffix}"
                headers = {"Range": "bytes=18-25"}
                response: requests.Response = requests.get(dumbfile, headers=headers)
                npartitions: int = int.from_bytes(response.content, "little")

                # Check if all parts are there
                files = []
                for d in range(npartitions):
                    partial_file = (
                        f"{path}/{filename}.{str(d).zfill(4)}.part{path_suffix}"
                    )
                    files.append(partial_file)
            else:
                # Get all files in the directory
                file = pathlib.Path(file)
                filename = pathlib.Path(filename)

                # It is expected that the file is a multi-part file in the same directory
                dumbfile: pathlib.Path = file.resolve().parent / (
                    filename.stem + f".0000.part{path_suffix}"
                )

                # check how many parts are there
                with open(dumbfile, "rb") as f:
                    f.seek(18)
                    npartitions: int = int.from_bytes(f.read(8), "little")

                # Check if all parts are there
                files = []
                for d in range(npartitions):
                    partial_file = filename.with_suffix(
                        f".{str(d).zfill(4)}.part.{path_suffix}"
                    )
                    if partial_file.exists():
                        files.append(partial_file)  # Add the file to the list
                    else:
                        raise FileNotFoundError(f"Missing partial file: {partial_file}")

        else:
            files = file

    return files


def sort_columns_add_geometry(metadata):
    """Sort the columns of a metadata DataFrame.
    Also, convert the "stac:centroid" column to a geometry column.

    Args:
        metadata (pd.DataFrame): The metadata DataFrame.

    Returns:
        pd.DataFrame: The metadata DataFrame with sorted columns.
    """
    if "stac:centroid" in metadata.columns:
        metadata = gpd.GeoDataFrame(
            data=metadata,
            geometry=metadata["stac:centroid"].apply(shapely.wkt.loads),
            crs="EPSG:4326",
        )
    columns = metadata.columns
    prefixes = ["internal:", "tortilla:", "stac:", "rai:"]
    sorted_columns = [
        col for prefix in prefixes for col in columns if col.startswith(prefix)
    ]
    rest = [col for col in columns if col not in sorted_columns and col != "geometry"]
    columns = sorted_columns + rest + (["geometry"] if "geometry" in columns else [])
    return metadata[columns]



import pathlib
from typing import List, Union

from tacoreader.load import load_metadata

try:
    from mdutils.mdutils import MdUtils
except ImportError:
    raise ImportError("Please install the mdutils package with: pip install mdutils")


def read_datacard(
    file: Union[str, pathlib.Path, List[pathlib.Path], List[str]],
    outfile: Union[str, pathlib.Path],
) -> pathlib.Path:

    # Create a JSON file
    taco_object: dict = load_metadata(file)

    # Convert the output file to a pathlib.Path object
    if isinstance(outfile, str):
        outfile = pathlib.Path(outfile)

    # Create the README.md file
    md_file = MdUtils(file_name=outfile)

    # --- YAML Header ---
    md_file.new_line("---")
    md_file.new_line("license:")
    for item in taco_object["licenses"]:
        md_file.new_line(f"  - {item}")
    md_file.new_line("language:")
    md_file.new_line("- en")
    if taco_object["keywords"]:
        md_file.new_line("tags:")
        for tag in taco_object["keywords"]:
            md_file.new_line(f"  - {tag}")
    md_file.new_line(f'pretty_name: {taco_object["id"]}')
    md_file.new_line("---")

    # --- Title and Description ---
    md_file.new_header(level=1, title=taco_object["id"])
    if taco_object["title"]:
        md_file.new_line(f'**{taco_object["title"]}**', bold_italics_code="b")
    md_file.new_paragraph(taco_object["description"])

    # --- Code Snippet ---
    md_file.new_header(level=2, title="🌮 TACO Snippet")
    md_file.new_paragraph("Load this dataset using the `tacoreader` library.")
    md_file.new_line("```python")
    md_file.new_line("import tacoreader")
    md_file.new_line("dataset = tacoreader.load('...')")
    md_file.new_line("```")

    # example in R
    md_file.new_line("\n")
    md_file.new_paragraph("Or in R:")
    md_file.new_line("```r")
    md_file.new_line("library(tacoreader)")
    md_file.new_line("dataset <- tacoreader::load('...')")
    md_file.new_line("```")

    # --- Sensor Information ---
    if taco_object["optical_data"]:
        md_file.new_header(level=2, title="🛰️ Sensor Information")
        md_file.new_paragraph(
            f'The sensor related to the dataset: **{taco_object["optical_data"]["sensor"]}**'
        )

    # --- Task Information ---
    md_file.new_header(level=2, title="🎯 Task")
    md_file.new_paragraph(
        f'The task associated with this dataset: **{taco_object["task"]}**'
    )

    # --- Raw Repository Link ---
    if taco_object["raw_link"]:
        md_file.new_header(level=2, title="📂 Original Data Repository")
        md_file.new_paragraph(
            f'Source location of the raw data:**{md_file.new_inline_link(link=taco_object["raw_link"]["href"])}**'
        )

    # --- Discussion Link ---
    if taco_object["discuss_link"]:
        md_file.new_header(level=2, title="💬 Discussion")
        md_file.new_paragraph(
            f'Insights or clarifications about the dataset: **{md_file.new_inline_link(link=taco_object["discuss_link"]["href"])}**'
        )

    # --- Split Strategy ---
    if taco_object["split_strategy"]:
        md_file.new_header(level=2, title="🔀 Split Strategy")
        md_file.new_paragraph(
            f'How the dataset is divided for training, validation, and testing: **{taco_object["split_strategy"]}**'
        )

    # --- Scientific Publications ---
    if taco_object["scientific"]["publications"]:
        md_file.new_header(level=2, title="📚 Scientific Publications")
        md_file.new_paragraph("Publications that reference or describe the dataset.")
        for idx, pub in enumerate(taco_object["scientific"]["publications"], start=1):
            # Add the publication information
            md_file.new_paragraph(f"### Publication {idx:02d}")
            md_file.new_line("- **DOI**: " + md_file.new_inline_link(link=pub["doi"]))
            md_file.new_line("- **Summary**: " + pub["summary"])
            md_file.new_line("- **BibTeX Citation**:")
            md_file.new_line("```bibtex")
            md_file.new_line(pub["citation"].strip("\n"))
            md_file.new_line("```")
            md_file.new_line("\n")

    # --- Data Providers ---
    if taco_object["providers"]:
        md_file.new_header(level=2, title="🤝 Data Providers")
        md_file.new_paragraph(
            "Organizations or individuals responsible for the dataset."
        )

        # Define table headers
        table_headers = ["**Name**", "**Role**", "**URL**"]
        table_data = [table_headers]

        # Populate table with provider data
        for provider in taco_object["providers"]:
            table_data.append(
                [
                    provider["name"] or "N/A",
                    (
                        ", ".join(provider["roles"])
                        if isinstance(provider["roles"], list)
                        else provider["roles"] or "N/A"
                    ),
                    (
                        md_file.new_inline_link(link=provider["links"][0]["href"])
                        if provider["links"]
                        else "N/A"
                    ),
                ]
            )

        # Flatten the list for Markdown formatting
        flat_table_data = [cell for row in table_data for cell in row]

        # Create the table
        md_file.new_table(
            columns=3, rows=len(table_data), text=flat_table_data, text_align="left"
        )

    # --- Curators ---
    if taco_object["curators"]:
        md_file.new_header(level=2, title="🧑‍🔬 Curators")
        md_file.new_paragraph(
            "Responsible for structuring the dataset in the TACO format."
        )

        # Define table headers
        table_headers = ["**Name**", "**Organization**", "**URL**"]
        table_data = [table_headers]

        # Populate table with curator data
        for curator in taco_object["curators"]:
            table_data.append(
                [
                    curator["name"] or "N/A",
                    curator["organization"] or "N/A",
                    (
                        md_file.new_inline_link(link=curator["links"][0]["href"])
                        if curator["links"]
                        else "N/A"
                    ),
                ]
            )

        # Flatten the list for Markdown formatting
        flat_table_data = [cell for row in table_data for cell in row]
        # Create the table
        md_file.new_table(
            columns=3, rows=len(table_data), text=flat_table_data, text_align="left"
        )

    # --- Labels ---
    if taco_object["labels"]:
        md_file.new_header(level=2, title="🏷️ Labels")
        md_file.new_paragraph(taco_object["labels"]["label_description"])
        table_headers = ["**Name**", "**Category**", "**Description**"]
        table_data = [table_headers]
        for item in taco_object["labels"]["label_classes"]:
            table_data.append(
                [
                    item["name"] or "N/A",
                    (
                        str(item["category"]) if item["category"] is not None else "N/A"
                    ),  # Zero is a valid category
                    item["description"] or "N/A",
                ]
            )

        # Flatten the list for Markdown formatting
        flat_table_data = [cell for row in table_data for cell in row]
        md_file.new_table(
            columns=3, rows=len(table_data), text=flat_table_data, text_align="left"
        )

    # --- Optical Bands ---
    if taco_object["optical_data"]:
        md_file.new_header(level=2, title="🌈 Optical Bands")
        md_file.new_paragraph("Spectral bands related to the sensor.")
        table_headers = [
            "**Name**",
            "**Common Name**",
            "**Description**",
            "**Center Wavelength**",
            "**Full Width Half Max**",
            "**Index**",
        ]
        table_data = [table_headers]

        for item in taco_object["optical_data"]["bands"]:
            table_data.append(
                [
                    item["name"] or "N/A",
                    item["common_name"] or "N/A",
                    item["description"] or "N/A",
                    item["center_wavelength"] or "N/A",
                    item["full_width_half_max"] or "N/A",
                    str(item["index"]) if item["index"] is not None else "N/A",
                ]
            )

        # Flatten the list for Markdown formatting
        flat_table_data = [cell for row in table_data for cell in row]
        md_file.new_table(
            columns=6, rows=len(table_data), text=flat_table_data, text_align="left"
        )

    # Export the data to the output file
    file = md_file.get_md_text().replace("\n\n\n  \n", "").replace("  \n", "\n")
    with open(outfile, "w") as f:
        f.write(file)

    return outfile



import random
from typing import Dict

import pandas as pd


def tortilla_message() -> str:
    """Get a random tortilla message"""

    tortilla_messages = [
        "Making a tortilla",
        "Making a tortilla 🫓",
        "Cooking a tortilla",
        "Making a tortilla 🫓",
        "Working on a tortilla",
        "Working on a tortilla 🫓",
        "Rolling out a tortilla",
        "Rolling out a tortilla 🫓",
        "Baking a tortilla",
        "Baking a tortilla 🫓",
        "Grilling a tortilla",
        "Grilling a tortilla 🫓",
        "Toasting a tortilla",
        "Toasting a tortilla 🫓",
    ]

    # Randomly accessing a message
    random_message = random.choice(tortilla_messages)
    return random_message


def build_simplified_range_header(df: pd.DataFrame) -> Dict[str, str]:
    """Build a simplified range header from a DataFrame of byte ranges
    That means that consecutive byte ranges will be merged into a single range.

    Args:
        df (pd.DataFrame): A DataFrame with columns "offset" and "length"
            that represent the byte ranges.

    Returns:
        Dict[str, str]: A dictionary with the "Range" header
    """

    # Create a list to hold the simplified byte ranges
    simplified_ranges = []

    # Get the initial offset and length
    current_start = df.iloc[0]["tortilla:offset"]
    current_end = current_start + df.iloc[0]["tortilla:length"]

    # Iterate over the rows starting from the second row
    for i in range(1, len(df)):
        row = df.iloc[i]
        next_start = row["tortilla:offset"]
        next_end = next_start + row["tortilla:length"]

        # If the current range and the next range are consecutive, merge them
        if next_start == current_end:
            current_end = next_end
        else:
            # Otherwise, add the current range to the list and start a new one
            simplified_ranges.append((current_start, current_end))
            current_start = next_start
            current_end = next_end

    # Append the final range
    simplified_ranges.append((current_start, current_end))

    # Create the range header
    range_header = ",".join(
        [f"bytes={start}-{end-1}" for start, end in simplified_ranges]
    )
    headers = {"Range": range_header}

    return headers


def human2bytes(size_str: str) -> int:
    """
    Converts a human-readable size string (e.g., "100MB") into bytes.
    Supported units: KB, MB, GB, TB, PB.
    """
    units = {"KB": 10**3, "MB": 10**6, "GB": 10**9, "TB": 10**12, "PB": 10**15}
    size_str = size_str.strip().upper()

    for unit, multiplier in units.items():
        if size_str.endswith(unit):
            try:
                value = float(size_str[: -len(unit)].strip())
                return int(value * multiplier)
            except ValueError:
                raise ValueError(f"Invalid size value in '{size_str}'.")
    raise ValueError(
        f"Unsupported unit in '{size_str}'. Supported units are: {', '.join(units.keys())}."
    )



datasets = {
    "tacofoundation/sen2naipv2-unet": [
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-unet.0000.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-unet.0001.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-unet.0002.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-unet.0003.part.taco",
    ],
    "tacofoundation/sen2naipv2-histmatch": [
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-histmatch.0000.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-histmatch.0001.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-histmatch.0002.part.taco",
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-histmatch.0003.part.taco",
    ],
    "tacofoundation/sen2naipv2-crosssensor": [
        "https://huggingface.co/datasets/tacofoundation/SEN2NAIPv2/resolve/main/sen2naipv2-crosssensor.taco",
    ],
}
